from __future__ import annotations

import logging
import pathlib
import time

import anywidget
import duckdb
import pyarrow as pa
import traitlets

from ._util import (
    arrow_table_from_dataframe_protocol,
    arrow_table_from_ipc,
    get_columns,
    has_pycapsule_stream_interface,
    is_arrow_ipc,
    is_dataframe_api_obj,
    is_polars,
    table_to_ipc,
)

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

SLOW_QUERY_THRESHOLD = 5000


class Widget(anywidget.AnyWidget):
    """An anywidget for displaying tabular data in a table."""

    _esm = pathlib.Path(__file__).parent / "widget.js"
    _table_name = traitlets.Unicode().tag(sync=True)
    _columns = traitlets.List(traitlets.Unicode()).tag(sync=True)

    # The SQL query for the current data (read-only)
    sql = traitlets.Unicode().tag(sync=True)

    # Where data cube indexes should be created
    data_cube_schema = traitlets.Unicode("quak").tag(sync=True)

    def __init__(self, data, *, table: str = "df"):
        if isinstance(data, duckdb.DuckDBPyConnection):
            conn = data
        else:
            conn = duckdb.connect(":memory:")
            if is_polars(data):
                # FIXME: special case pl.DataFrame for now until DuckDB
                # supports `[string,bytes]_view` Arrow data types
                # see: https://github.com/manzt/quak/issues/41
                # Polars .to_arrow() method will cast to non-view array types for us
                arrow_table = data.to_arrow()
            elif has_pycapsule_stream_interface(data):
                # NOTE: for now we materialize the input into an in-memory Arrow table,
                # so that we can perform repeated queries on that. In the future, it may
                # be better to keep this Arrow stream non-materalized in Python and
                # create a new DuckDB table from the stream.
                # arrow_table = pa.RecordBatchReader.from_stream(data)
                arrow_table = pa.table(data)
            elif is_arrow_ipc(data):
                arrow_table = arrow_table_from_ipc(data)
            elif is_dataframe_api_obj(data):
                arrow_table = arrow_table_from_dataframe_protocol(data)
            else:
                raise ValueError(
                    "input must be a DuckDB connection, DataFrame-like, an Arrow IPC "
                    "table, or an Arrow object exporting the Arrow C Stream interface."
                )
            conn.register(table, arrow_table)
        self._conn = conn
        super().__init__(
            _table_name=table,
            _columns=get_columns(conn, table),
            sql=f'SELECT * FROM "{table}"',
        )
        self.on_msg(self._handle_custom_msg)

    def _handle_custom_msg(self, data: dict, buffers: list):
        logger.debug(f"{data=}, {buffers=}")

        start = time.time()

        uuid = data["uuid"]
        sql = data["sql"]
        command = data["type"]
        try:
            if command == "arrow":
                result = self._conn.query(sql).arrow()
                buf = table_to_ipc(result)
                self.send({"type": "arrow", "uuid": uuid}, buffers=[buf])
            elif command == "exec":
                self._conn.execute(sql)
                self.send({"type": "exec", "uuid": uuid})
            elif command == "json":
                result = self._conn.query(sql).df()
                json = result.to_dict(orient="records")
                self.send({"type": "json", "uuid": uuid, "result": json})
            else:
                raise ValueError(f"Unknown command {command}")
        except Exception as e:
            logger.exception("Error processing query")
            self.send({"error": str(e), "uuid": uuid})

        total = round((time.time() - start) * 1_000)
        if total > SLOW_QUERY_THRESHOLD:
            logger.warning(f"DONE. Slow query {uuid} took {total} ms.\n{sql}")
        else:
            logger.info(f"DONE. Query {uuid} took {total} ms.\n{sql}")

    def data(self) -> duckdb.DuckDBPyRelation:
        """Return the current SQL as a DuckDB relation."""
        return self._conn.query(self.sql)



"""A Hatchling plugin to build the quak frontend."""

import os
import pathlib
import subprocess

from hatchling.builders.hooks.plugin.interface import BuildHookInterface

ROOT = pathlib.Path(__file__).parent / ".."


class QuakBuildHook(BuildHookInterface):
    """Hatchling plugin to build the quak frontend."""

    PLUGIN_NAME = "quak"

    def initialize(self, version: str, build_data: dict) -> None:
        """Initialize the plugin."""
        if os.getenv("SKIP_DENO_BUILD", "0") == "1":
            # Skip the build if the environment variable is set
            # Useful in CI/CD pipelines
            return

        if not (ROOT / "src/quak/widget.js").exists():
            subprocess.check_call(["deno", "task", "build", "--bundle"], cwd=ROOT)



from __future__ import annotations

import sys
import typing

DataFrameObject = typing.Any

if typing.TYPE_CHECKING:
    import duckdb
    import polars as pl
    import pyarrow as pa


def has_pycapsule_stream_interface(obj: object) -> bool:
    """
    Check if an object implements the Arrow C Stream Arrow via the PyCapsule Interface.

    https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html
       has an Arrow C Stream
    """
    return hasattr(obj, "__arrow_c_stream__")


def is_dataframe_api_obj(obj: object) -> DataFrameObject:
    """Check if an object has a dataframe API."""
    method = getattr(obj, "__dataframe__", None)
    return callable(method)


def is_arrow_ipc(x: object) -> typing.TypeGuard[bytes | memoryview]:
    """Check if an object is an Arrow IPC message."""
    if not isinstance(x, (bytes, memoryview)):
        return False
    magic = b"ARROW1"
    if isinstance(x, memoryview):
        x = x.tobytes()
    return x[:6] == magic


# Copied from Altair
# https://github.com/vega/altair/blob/18a2c3c237014591d172284560546a2f0ac1a883/altair/utils/data.py#L343
def arrow_table_from_dataframe_protocol(dflike: DataFrameObject) -> pa.lib.Table:
    """Convert a DataFrame-like object to a pyarrow Table."""
    import pyarrow as pa
    import pyarrow.interchange as pi

    # First check if the dataframe object has a method to convert to arrow.
    # Give this preference over the pyarrow from_dataframe function
    # since the object
    # has more control over the conversion, and may have broader compatibility.
    # This is the case for Polars, which supports Date32 columns in
    # direct conversion
    # while pyarrow does not yet support this type in from_dataframe
    for convert_method_name in ("arrow", "to_arrow", "to_arrow_table"):
        convert_method = getattr(dflike, convert_method_name, None)
        if callable(convert_method):
            result = convert_method()
            if isinstance(result, pa.Table):
                return result

    return pi.from_dataframe(dflike)  # type: ignore[no-any-return]


def arrow_table_from_ipc(data: bytes | memoryview) -> pa.lib.Table:
    """Convert an Arrow IPC message to a pyarrow Table."""
    import io

    import pyarrow.feather as feather

    return feather.read_table(io.BytesIO(data))


def table_to_ipc(table: pa.lib.Table) -> memoryview:
    """Convert a pyarrow Table to an Arrow IPC message."""
    import io

    import pyarrow.feather as feather

    sink = io.BytesIO()
    feather.write_feather(table, sink, compression="uncompressed")
    return sink.getbuffer()


def record_batch_to_ipc(record_batch: pa.lib.RecordBatch) -> memoryview:
    """Convert a pyarrow RecordBatch to an Arrow IPC message."""
    import pyarrow as pa

    table = pa.Table.from_batches([record_batch], schema=record_batch.schema)
    return table_to_ipc(table)


def get_columns(conn: duckdb.DuckDBPyConnection, table_name: str) -> list[str]:
    """Determine the column names of a table in a database."""
    result = conn.execute(f"""
    SELECT column_name
    FROM information_schema.columns
    WHERE table_name = '{table_name}'
    """)
    rows = result.fetchall()
    return [row[0] for row in rows]


def is_polars(obj: object) -> typing.TypeGuard[pl.DataFrame]:
    """Check if an object is a Polars DataFrame."""
    polars = sys.modules.get("polars")
    return polars is not None and isinstance(obj, polars.DataFrame)
